package com.novayre.jidoka.robot;

import java.util.Map;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import org.apache.commons.lang3.StringUtils;
import org.apache.commons.lang3.exception.ExceptionUtils;
import org.openqa.selenium.By;
import org.openqa.selenium.WebElement;

import com.novayre.jidoka.browser.api.EBrowsers;
import com.novayre.jidoka.browser.api.IWebBrowserSupport;
import com.novayre.jidoka.client.api.IJidokaServer;
import com.novayre.jidoka.client.api.IRobot;
import com.novayre.jidoka.client.api.JidokaFactory;
import com.novayre.jidoka.client.api.annotations.Robot;
import com.novayre.jidoka.client.api.exceptions.JidokaFatalException;
import com.novayre.jidoka.client.api.multios.IClient;

/**
 * Context Robot.
 * 
 * @author Jidoka
 *
 */
@Robot
public class ContextRobot implements IRobot {

	/**
	 * Robot input parameter to indicate if the context must be reset or not.
	 */
	private static final String RESET_CONTEXT_PARAM = "reset";

	/**
	 * Regular expression to extract the version of Internet Explorer web driver.
	 */
	private static final String REGEX_IE = "\\d+[.]\\d+[.]\\d+";

	/**
	 * Regular expression to extract the version of Chrome web driver.
	 */
	private static final String REGEX_CHROME = "(\\d|[.])+";

	/**
	 * XPath expression to get the information related to the Internet Explorer web
	 * driver version.
	 */
	private static final String XPATH_IE_DRIVER = "(//h2[text()='The Internet Explorer Driver Server'])/following::p[2]";

	/**
	 * XPath expression to get the information related to the Chrome web driver
	 * version.
	 */
	private static final String XPATH_CHROME_DRIVER = "//span[contains(text(),'Current stable')]/following-sibling::a";

	/**
	 * XPath expression to get the information related to the Firefox web driver
	 * version.
	 */
	private static final String XPATH_FIREFOX_DRIVER = "(//div[@class='release-header'])[1]//a";

	/**
	 * URL for downloading Chrome driver: chromedriver.exe
	 */
	private static final String URL_CHROME = "https://sites.google.com/a/chromium.org/chromedriver/";

	/**
	 * URL for downloading IE driver: IEDriverServer.exe. This driver is provided by
	 * Selenium.
	 */
	private static final String URL_IE = "http://www.seleniumhq.org/download/";

	/**
	 * URL for downloading the Firefox driver: geckodriver.exe
	 */
	private static final String URL_FIREFOX = "https://github.com/mozilla/geckodriver/releases";

	/**
	 * Server.
	 */
	private IJidokaServer<DriverContext> server;

	/**
	 * Browser module.
	 */
	private IWebBrowserSupport browser;

	/**
	 * Variable to store the current version of ChromeDriver retrieved from the web.
	 */
	private String currentChromeDriverVersion;

	/**
	 * Variable to store the current version of Internet Explorer Driver retrieved
	 * from the web.
	 */
	private String currentIEDriverVersion;

	/**
	 * Variable to store the current version of Firefox Driver retrieved from the
	 * web.
	 */
	private String currentFirefoxDriverVersion;

	@SuppressWarnings("unchecked")
	@Override
	public boolean startUp() throws Exception {
		// Init Jidoka tools.
		server = (IJidokaServer<DriverContext>) JidokaFactory.getServer();
		IClient client = IClient.getInstance(this);
		browser = IWebBrowserSupport.getInstance(this, client);
		return IRobot.super.startUp();
	}

	/**
	 * Action 'Start'.
	 * <p>
	 * Starts the robot execution.
	 * 
	 * @throws Exception any exception
	 */
	public void start() {

		try {
			server.setNumberOfItems(3);

			// Browser
			browser.setBrowserType(EBrowsers.CHROME);
			browser.initBrowser();
		} catch (Exception e) {
			throw new JidokaFatalException("Error initializing the robot");
		}
	}

	/**
	 * Checks the Chrome web driver version.
	 */
	public void checkChrome() {

		try {
			WebElement we = checkDriver(URL_CHROME, XPATH_CHROME_DRIVER);

			currentChromeDriverVersion = extractDriverVersion(we, REGEX_CHROME);

			infoDriver(currentChromeDriverVersion);
		} catch (Exception e) {
			throw new JidokaFatalException("Error checking CHROME version");
		}
	}

	/**
	 * Checks the Internet Explorer web driver version.
	 */
	public void checkIE() {

		try {
			WebElement we = checkDriver(URL_IE, XPATH_IE_DRIVER);

			currentIEDriverVersion = extractDriverVersion(we, REGEX_IE);

			infoDriver(currentIEDriverVersion);
		} catch (Exception e) {
			throw new JidokaFatalException("Error checking IE version");
		}
	}

	/**
	 * Checks the Firefox web driver version.
	 */
	public void checkFirefox() {

		try {
			WebElement we = checkDriver(URL_FIREFOX, XPATH_FIREFOX_DRIVER);

			currentFirefoxDriverVersion = extractDriverVersion(we, null);

			infoDriver(currentFirefoxDriverVersion);
		} catch (Exception e) {
			throw new JidokaFatalException("Error checking FIREFOX version");
		}
	}

	/**
	 * Checks the URL passed by parameter, access it and evaluates the XPath
	 * expression specified to get an HTML containing the information of the web
	 * driver.
	 * 
	 * @param url             URL to access to
	 * @param xpathExpression expression to evaluate
	 * @return the {@link WebElement} containing the driver information to be
	 *         processed later
	 */
	private WebElement checkDriver(String url, String xpathExpression) {

		// Navigate to the driver page
		browser.navigate(url);

		// WebElement containing the driver version
		return browser.getElement(By.xpath(xpathExpression));
	}

	/**
	 * Extracts the web driver version contained in the {@link WebElement}
	 * specified. If a regular expression is specified, it is applied to get the web
	 * driver version.
	 * 
	 * @param we    the {@link WebElement} with the web driver information
	 * @param regex the regular expression to evaluate
	 * @return the extracted version
	 */
	private String extractDriverVersion(WebElement we, String regex) {

		// If no regular expression is present, return
		if (regex == null) {
			return we.getText();
		}

		// A regular expression must be evaluated
		server.info(
				String.format("Serching web driver using the regular expression %s on text %s", regex, we.getText()));

		// Search with regular expression
		Pattern p = Pattern.compile(regex);
		Matcher m = p.matcher(we.getText());
		if (m.find()) {
			// The version extracted from the method parameter 'we'
			return m.group();
		}

		// Nothing found using the regular expression
		throw new JidokaFatalException(String.format("Web driver not found using the regex %s", regex));
	}

	/**
	 * Print an INFO message to the execution log to show the driver version.
	 * 
	 * @param driverVersion the driver version
	 */
	private void infoDriver(String driverVersion) {

		server.info(String.format("The current web driver version is %s", driverVersion));
	}

	/**
	 * Action 'Compare Versions with Context'.
	 * <p>
	 * Checks all drivers versions and compares them to those saved previously in
	 * the context.
	 */
	public void compareDriversVersions() {

		try {
			// Get the context
			DriverContext driverContext = server.getPersistentContextValue();

			if (driverContext == null) {
				server.info("The context does not exist. It was reset or this is is the first time the robot is executed");
				driverContext = new DriverContext();
			} else {

				server.info("Context retrieved successfully");

				server.info(String.format("IEDriverVersion %s", driverContext.getLastIEDriverVersion()));
				server.info(String.format("ChromeDriverVersion %s", driverContext.getLastChromeDriverVersion()));
				server.info(String.format("FirefoxDriverVersion %s", driverContext.getLastFirefoxDriverVersion()));
			}

			// IE
			server.setCurrentItem(1, "Internet Explorer");
			compareVersion(driverContext.getLastIEDriverVersion(), currentIEDriverVersion);
			driverContext.setLastIEDriverVersion(currentIEDriverVersion);

			// Chrome
			server.setCurrentItem(2, "Chrome");
			compareVersion(driverContext.getLastChromeDriverVersion(), currentChromeDriverVersion);
			driverContext.setLastChromeDriverVersion(currentChromeDriverVersion);

			// Firefox
			server.setCurrentItem(3, "Firefox");
			compareVersion(driverContext.getLastFirefoxDriverVersion(), currentFirefoxDriverVersion);
			driverContext.setLastFirefoxDriverVersion(currentFirefoxDriverVersion);

			// Set the context
			server.setPersistentContextValue(driverContext);
		} catch (Exception e) {
			throw new JidokaFatalException("Error comparing drivers version");
		}
	}

	/**
	 * Method to compare the value obtained from the driver web to the values
	 * retrieved from the context.
	 * <p>
	 * <ul>
	 * <li>If there is some differences, the item will be set to WARN.</li>
	 * <li>If there is not any difference, the item will be set to OK.</li>
	 * <ul>
	 * 
	 * @param previousVersion version stored in the context
	 * @param currentVersion  version obtained from the web
	 */
	private void compareVersion(String previousVersion, String currentVersion) {

		if (currentVersion.equals(previousVersion)) {

			server.info(String.format("The web driver has not changed: '%s'", currentVersion));
			server.setCurrentItemResultToOK(currentVersion);
		} else {

			String wrnMsg = String.format("The web driver has changed. Previous version was '%s', New version is %s'",
					previousVersion, currentVersion);
			server.info(wrnMsg);
			server.setCurrentItemResultToWarn(wrnMsg);
		}
	}

	/**
	 * Action 'End'.
	 * <p>
	 * Resets the context in case the robot execution indicates so.
	 */
	public void end() {

		Map<String, String> parameters = server.getParameters();

		boolean reset = Boolean.parseBoolean(parameters.get(RESET_CONTEXT_PARAM));

		if (reset) {
			server.resetPersistenContextValue();

			server.info("The context was reset");
		}

		// Close the browser
		browser.close();
	}

	/**
	 * 
	 * @see com.novayre.jidoka.client.api.IRobot#cleanUp()
	 */
	@Override
	public String[] cleanUp() throws Exception {

		// Close the browser
		if (browser != null) {
			browser.close();
		}

		return new String[0];
	}

	/**
	 * Manage exception.
	 *
	 * @param action    the action
	 * @param exception the exception
	 * @return the string
	 * @throws Exception the exception
	 */
	@Override
	public String manageException(String action, Exception exception) throws Exception {

		// We get the message of the exception
		String errorMessage = ExceptionUtils.getRootCause(exception).getMessage();

		// We send a screenshot to the log so the user can see the screen in the moment
		// of the error
		// This is a very useful thing to do
		server.sendScreen("Screenshot at the moment of the error");

		// If we have a FatalException we should abort the execution.
		if (ExceptionUtils.indexOfThrowable(exception, JidokaFatalException.class) >= 0) {

			server.error(StringUtils.isBlank(errorMessage) ? "Fatal error" : errorMessage);
			return IRobot.super.manageException(action, exception);
		}

		server.warn("Unknown exception!");

		// If we have any other exception we must abort the execution, we don't know
		// what has happened

		return IRobot.super.manageException(action, exception);
	}

}
